# -*- coding: utf-8 -*-
#--
# threadmanager - A simple parallel processing thread manager
#++
# Copyright (c) 2009 Akinori MUSHA <knu@iDaemons.org>
#
# All rights reserved.  You can redistribute and/or modify it under the same
# terms as Ruby.
#
# == Synopsis
#
#   require 'threadmanager'
#
#   tm = ThreadManager.new($MAX_PROCESSES)
#
#   all_data.each { |data|
#     th = tm.start {
#       # ... do some work with data in the thread ...
#     }
#   }
#

class ThreadManager
  VERSION = '0.7.5'

  def initialize(concurrency)
    @concurrency = concurrency
    @processes = {}
    @on_finish = {}
  end

  attr_accessor :concurrency

  def start(ident = nil, &block)
    raise "Cannot start another process while you are in the child process" if in_child?
    while @processes.size >= @concurrency
      invoke_on_wait
      wait_one_child
    end
    wait_children
    th = Thread.start(&block)
    @processes[th] = ident
    invoke_on_start(th, ident)
    th
  end

  def in_child?
    @processes.include?(Thread.current)
  end

  def on_start(&code)
    @on_start = code
  end

  def on_finish(th = nil, &code)
    @on_finish[th] = code
  end

  def on_wait(period = nil, &code)
    @on_wait = code
    @on_wait_period = period
  end

  def wait_all_children
    until @processes.empty?
      invoke_on_wait
      wait_one_child
    end
  end

  def self.run(concurrency, &block)
    obj = new(concurrency)
    block.call(obj)
    obj.wait_all_children
  end

  private

  def wait_children
    return if @processes.empty?
    while wait_one_child(0)
    end
  end

  def wait_one_child(timeout = @on_wait_period)
    @processes.each_key { |th|
      if timeout
        th.join(timeout) or next
      else
        th.join
      end
      if @processes.key?(th)
        ident = @processes.delete(th)
        invoke_on_finish(th, ident)
        return th
      end
    }
    return nil
  end

  def invoke_on_start(th, ident)
    @on_start.call(th, ident) if @on_start
  end

  def invoke_on_finish(th, ident)
    code = @on_finish[th] || @on_finish[nil] or return nil
    code.call(th, ident)
  end

  def invoke_on_wait
    @on_wait.call if @on_wait
  end
end
